@@ -4,7 +4,6 @@ import os
 import subprocess
 from tempfile import TemporaryFile, NamedTemporaryFile
 import wave
-import audioop
 import sys
 
 try:
@@ -16,35 +15,85 @@ from .utils import (
     _fd_or_path_or_tempfile,
     db_to_float,
     ratio_to_db,
+    get_encoder_name,
+    audioop,
 )
 from .exceptions import (
     TooManyMissingFrames,
     InvalidDuration,
     InvalidID3TagVersion,
     InvalidTag,
+    CouldntDecodeError,
 )
 
 if sys.version_info >= (3, 0):
     basestring = str
     xrange = range
     StringIO = BytesIO
+    
+
+class ClassPropertyDescriptor(object):
+
+    def __init__(self, fget, fset=None):
+        self.fget = fget
+        self.fset = fset
+
+    def __get__(self, obj, klass=None):
+        if klass is None:
+            klass = type(obj)
+        return self.fget.__get__(obj, klass)()
+
+    def __set__(self, obj, value):
+        if not self.fset:
+            raise AttributeError("can't set attribute")
+        type_ = type(obj)
+        return self.fset.__get__(obj, type_)(value)
+
+    def setter(self, func):
+        if not isinstance(func, (classmethod, staticmethod)):
+            func = classmethod(func)
+        self.fset = func
+        return self    
+
+def classproperty(func):
+    if not isinstance(func, (classmethod, staticmethod)):
+        func = classmethod(func)
+
+    return ClassPropertyDescriptor(func)
+
 
 AUDIO_FILE_EXT_ALIASES = {
-    "m4a": "mp4"
+    "m4a": "mp4",
+    "wave": "wav",
 }
 
 
 class AudioSegment(object):
-
     """
-    Note: AudioSegment objects are immutable
+    AudioSegments are *immutable* objects representing segments of audio
+    that can be manipulated using python code.
 
-    slicable using milliseconds. for example: Get the first second of an mp3...
-
-    a = AudioSegment.from_mp3(mp3file)
-    first_second = a[:1000]
+    AudioSegments are slicable using milliseconds.
+    for example:
+        a = AudioSegment.from_mp3(mp3file)
+        first_second = a[:1000] # get the first second of an mp3
+        slice = a[5000:10000] # get a slice from 5 to 10 seconds of an mp3
     """
-    ffmpeg = 'ffmpeg'
+    converter = get_encoder_name()  # either ffmpeg or avconv
+
+    # TODO: remove in 1.0 release
+    # maintain backwards compatibility for ffmpeg attr (now called converter)
+    @classproperty
+    def ffmpeg(cls):
+        return cls.converter
+    
+    @ffmpeg.setter
+    def ffmpeg(cls, val):
+        cls.converter = val
+
+    DEFAULT_CODECS = {
+        "ogg": "libvorbis"
+    }
 
     def __init__(self, data=None, *args, **kwargs):
         if kwargs.get('metadata', False):
@@ -107,15 +156,41 @@ class AudioSegment(object):
         missing_frames = (expected_length - len(data)) // self.frame_width
         if missing_frames:
             if missing_frames > self.frame_count(ms=2):
-                raise TooManyMissingFrames("You should never be filling in "
-                                           "   more than 2 ms with silence here, missing frames: %s" %
-                                           missing_frames)
+                raise TooManyMissingFrames(
+                    "You should never be filling in "
+                    "   more than 2 ms with silence here, "
+                    "missing frames: %s" % missing_frames)
             silence = audioop.mul(data[:self.frame_width],
                                   self.sample_width, 0)
             data += (silence * missing_frames)
 
         return self._spawn(data)
 
+    def get_sample_slice(self, start_sample=None, end_sample=None):
+        """
+        Get a section of the audio segment by sample index.
+
+        NOTE: Negative indices do *not* address samples backword
+        from the end of the audio segment like a python list.
+        This is intentional.
+        """
+        max_val = self.frame_count()
+
+        def bounded(val, default):
+            if val is None:
+                return default
+            if val < 0:
+                return 0
+            if val > max_val:
+                return max_val
+            return val
+
+        start_i = bounded(start_sample, 0) * self.frame_width
+        end_i = bounded(end_sample, max_val) * self.frame_width
+
+        data = self._data[start_i:end_i]
+        return self._spawn(data)
+
     def __add__(self, arg):
         if isinstance(arg, AudioSegment):
             return self.append(arg, crossfade=0)
@@ -134,12 +209,13 @@ class AudioSegment(object):
         If the argument is an AudioSegment, overlay the multiplied audio
         segment.
 
-
         If it's a number, just use the string multiply operation to repeat the
-        audio so the following would return an AudioSegment that contains the
+        audio.
+
+        The following would return an AudioSegment that contains the
         audio of audio_seg eight times
 
-        audio_seg * 8
+        `audio_seg * 8`
         """
         if isinstance(arg, AudioSegment):
             return self.overlay(arg, position=0, loop=True)
@@ -148,10 +224,10 @@ class AudioSegment(object):
 
     def _spawn(self, data, overrides={}):
         """
-        Creates a new audio segment using the meta data from the current one
+        Creates a new audio segment using the metadata from the current one
         and the data passed in. Should be used whenever an AudioSegment is
-        being returned by an operation that alters the current one, since
-        AudioSegment objects are immutable.
+        being returned by an operation that would alters the current one,
+        since AudioSegment objects are immutable.
         """
         # accept lists of data chunks
         if isinstance(data, list):
@@ -184,6 +260,10 @@ class AudioSegment(object):
         seg1 = seg1.set_frame_rate(frame_rate)
         seg2 = seg2.set_frame_rate(frame_rate)
 
+        sample_width = max(seg1.sample_width, seg2.sample_width)
+        seg1 = seg1.set_sample_width(sample_width)
+        seg2 = seg2.set_sample_width(sample_width)
+
         assert(len(seg1) == s1_len)
         assert(len(seg2) == s2_len)
 
@@ -198,17 +278,41 @@ class AudioSegment(object):
 
     @classmethod
     def empty(cls):
-        return cls(b'', metadata={"channels": 1, "sample_width": 1, "frame_rate": 1, "frame_width": 1})
+        return cls(b'', metadata={
+            "channels": 1,
+            "sample_width": 1,
+            "frame_rate": 1,
+            "frame_width": 1
+        })
+
+    @classmethod
+    def silent(cls, duration=1000):
+        """
+        Generate a silent audio segment.
+        duration specified in milliseconds (default: 1000ms).
+        """
+        # lowest frame rate I've seen in actual use
+        frame_rate = 11025
+        frames = int(frame_rate * (duration / 1000.0))
+        data = b"\0\0" * frames
+        return cls(data, metadata={"channels": 1,
+                                   "sample_width": 2,
+                                   "frame_rate": frame_rate,
+                                   "frame_width": 2})
 
     @classmethod
     def from_file(cls, file, format=None):
+        orig_file = file
         file = _fd_or_path_or_tempfile(file, 'rb', tempfile=False)
 
         if format:
             format = AUDIO_FILE_EXT_ALIASES.get(format, format)
 
-        if format == 'wav':
-            return cls.from_wav(file)
+        if format == "wav" or (isinstance(orig_file, basestring) and orig_file.endswith(".wav")):
+            try:
+                return cls._from_safe_wav(file)
+            except:
+                pass
 
         input_file = NamedTemporaryFile(mode='wb', delete=False)
         input_file.write(file.read())
@@ -216,25 +320,28 @@ class AudioSegment(object):
 
         output = NamedTemporaryFile(mode="rb", delete=False)
 
-        ffmpeg_call = [cls.ffmpeg,
-                       '-y',  # always overwrite existing files
-                       ]
+        convertion_command = [cls.converter,
+                              '-y',  # always overwrite existing files
+                              ]
 
         # If format is not defined
-        # ffmpeg will detect it automatically
+        # ffmpeg/avconv will detect it automatically
         if format:
-            ffmpeg_call += ["-f", format]
+            convertion_command += ["-f", format]
 
-        ffmpeg_call += [
+        convertion_command += [
             "-i", input_file.name,  # input_file options (filename last)
             "-vn",  # Drop any video streams if there are any
             "-f", "wav",  # output options (filename last)
             output.name
         ]
 
-        subprocess.call(ffmpeg_call, stderr=open(os.devnull))
+        retcode = subprocess.call(convertion_command, stderr=open(os.devnull))
 
-        obj = cls.from_wav(output)
+        if retcode != 0:
+            raise CouldntDecodeError("Decoding failed. ffmpeg returned error code: {0}".format(retcode))
+
+        obj = cls._from_safe_wav(output)
 
         input_file.close()
         output.close()
@@ -257,6 +364,10 @@ class AudioSegment(object):
 
     @classmethod
     def from_wav(cls, file):
+        return cls.from_file(file, 'wav')
+
+    @classmethod
+    def _from_safe_wav(cls, file):
         file = _fd_or_path_or_tempfile(file, 'rb', tempfile=False)
         file.seek(0)
         return cls(data=file)
@@ -269,7 +380,8 @@ class AudioSegment(object):
             Path to destination audio file
 
         format (string)
-            Format for destination audio file. ('mp3', 'wav', 'ogg' or other ffmpeg supported files)
+            Format for destination audio file.
+            ('mp3', 'wav', 'ogg' or other ffmpeg/avconv supported files)
 
         codec (string)
             Codec used to encoding for the destination.
@@ -278,10 +390,11 @@ class AudioSegment(object):
             Bitrate used when encoding destination file. (128, 256, 312k...)
 
         parameters (string)
-            Aditional ffmpeg parameters
+            Aditional ffmpeg/avconv parameters
 
         tags (dict)
-            Set metadata information to destination files usually used as tags. ({title='Song Title', artist='Song Artist'})
+            Set metadata information to destination files
+            usually used as tags. ({title='Song Title', artist='Song Artist'})
 
         id3v2_version (string)
             Set ID3v2 version for tags. (default: '4')
@@ -301,8 +414,8 @@ class AudioSegment(object):
         wave_data.setnchannels(self.channels)
         wave_data.setsampwidth(self.sample_width)
         wave_data.setframerate(self.frame_rate)
-        # For some reason packing the wave header struct with a float in python 2
-        # doesn't throw an exception
+        # For some reason packing the wave header struct with
+        # a float in python 2 doesn't throw an exception
         wave_data.setnframes(int(self.frame_count()))
         wave_data.writeframesraw(self._data)
         wave_data.close()
@@ -313,49 +426,53 @@ class AudioSegment(object):
 
         output = NamedTemporaryFile(mode="w+b", delete=False)
 
-        # build ffmpeg command to export
-        ffmpeg_call = [self.ffmpeg,
-                       '-y',  # always overwrite existing files
-                       "-f", "wav", "-i", data.name,  # input options (filename last)
-                       ]
+        # build converter command to export
+        convertion_command = [
+            self.converter,
+            '-y',  # always overwrite existing files
+            "-f", "wav", "-i", data.name,  # input options (filename last)
+        ]
 
-        if format == "ogg" and codec is None:
-            ffmpeg_call.extend(["-acodec", "libvorbis"])
+        if codec is None:
+            codec = self.DEFAULT_CODECS.get(format, None)
 
         if codec is not None:
             # force audio encoder
-            ffmpeg_call.extend(["-acodec", codec])
+            convertion_command.extend(["-acodec", codec])
 
         if bitrate is not None:
-            ffmpeg_call.extend(["-b:a", bitrate])
+            convertion_command.extend(["-b:a", bitrate])
 
         if parameters is not None:
             # extend arguments with arbitrary set
-            ffmpeg_call.extend(parameters)
+            convertion_command.extend(parameters)
 
         if tags is not None:
             if not isinstance(tags, dict):
                 raise InvalidTag("Tags must be a dictionary.")
             else:
-                # Extend ffmpeg command with tags
-                [ffmpeg_call.extend(['-metadata', '{0}={1}'.format(k, v)])
-                 for k, v in tags.items()]
+                # Extend converter command with tags
+                # print(tags)
+                for key, value in tags.items():
+                    convertion_command.extend(
+                        ['-metadata', '{0}={1}'.format(key, value)])
 
                 if format == 'mp3':
+                    # set id3v2 tag version
                     if id3v2_version not in id3v2_allowed_versions:
                         raise InvalidID3TagVersion(
                             "id3v2_version not allowed, allowed versions: %s" % id3v2_allowed_versions)
-                    ffmpeg_call.extend([
+                    convertion_command.extend([
                         "-id3v2_version",  id3v2_version
                     ])
 
-        ffmpeg_call.extend([
+        convertion_command.extend([
             "-f", format, output.name,  # output options (filename last)
         ])
 
         # read stdin / write stdout
-        subprocess.call(ffmpeg_call,
-                        # make ffmpeg shut up
+        subprocess.call(convertion_command,
+                        # make converter shut up
                         stderr=open(os.devnull)
                         )
 
@@ -386,6 +503,25 @@ class AudioSegment(object):
         else:
             return float(len(self._data) // self.frame_width)
 
+    def set_sample_width(self, sample_width):
+        if sample_width == self.sample_width:
+            return self
+
+        data = self._data
+
+        if self.sample_width == 1:
+            data = audioop.bias(data, 1, -128)
+
+        if data:
+            data = audioop.lin2lin(data, self.sample_width, sample_width)
+
+        if sample_width == 1:
+            data = audioop.bias(data, 1, 128)
+
+        frame_width = self.channels * sample_width
+        return self._spawn(data, overrides={'sample_width': sample_width,
+                                            'frame_width': frame_width})
+
     def set_frame_rate(self, frame_rate):
         if frame_rate == self.frame_rate:
             return self
@@ -405,21 +541,32 @@ class AudioSegment(object):
             return self
 
         if channels == 2 and self.channels == 1:
-            fn = 'tostereo'
+            fn = audioop.tostereo
             frame_width = self.frame_width * 2
         elif channels == 1 and self.channels == 2:
-            fn = 'tomono'
+            fn = audioop.tomono
             frame_width = self.frame_width // 2
 
-        fn = getattr(audioop, fn)
         converted = fn(self._data, self.sample_width, 1, 1)
 
-        return self._spawn(data=converted, overrides={'channels': channels,
-                                                      'frame_width': frame_width})
+        return self._spawn(data=converted,
+                           overrides={
+                               'channels': channels,
+                               'frame_width': frame_width})
 
     @property
     def rms(self):
-        return audioop.rms(self._data, self.sample_width)
+        if self.sample_width == 1:
+            return self.set_sample_width(2).rms
+        else:
+            return audioop.rms(self._data, self.sample_width)
+
+    @property
+    def dBFS(self):
+        rms = self.rms
+        if not rms:
+            return - float("infinity")
+        return ratio_to_db(self.rms / self.max_possible_amplitude)
 
     @property
     def max(self):
@@ -441,7 +588,38 @@ class AudioSegment(object):
         return self._spawn(data=audioop.mul(self._data, self.sample_width,
                                             db_to_float(float(volume_change))))
 
-    def overlay(self, seg, position=0, loop=False):
+    def overlay(self, seg, position=0, loop=False, times=None):
+        """
+        Overlay the provided segment on to this segment starting at the
+        specificed position and using the specfied looping beahvior.
+
+        seg (AudioSegment):
+            The audio segment to overlay on to this one.
+
+        position (optional int):
+            The position to start overlaying the provided segment in to this
+            one.
+
+        loop (optional bool):
+            Loop seg as many times as necessary to match this segment's length.
+            Overrides loops param.
+
+        times (optional int):
+            Loop seg the specified number of times or until it matches this
+            segment's length. 1 means once, 2 means twice, ... 0 would make the
+            call a no-op
+        """
+
+        if loop:
+            # match loop=True's behavior with new times (count) mechinism.
+            times = -1
+        elif times is None:
+            # no times specified, just once through
+            times = 1
+        elif times == 0:
+            # it's a no-op, make a copy since we never mutate
+            return self._spawn(self._data)
+
         output = TemporaryFile()
 
         seg1, seg2 = AudioSegment._sync(self, seg)
@@ -456,27 +634,27 @@ class AudioSegment(object):
         pos = 0
         seg1_len = len(seg1)
         seg2_len = len(seg2)
-        while True:
+        while times:
             remaining = max(0, seg1_len - pos)
             if seg2_len >= remaining:
                 seg2 = seg2[:remaining]
                 seg2_len = remaining
-                loop = False
+                # we've hit the end, we're done looping (if we were) and this
+                # is our last go-around
+                times = 1
 
             output.write(audioop.add(seg1[pos:pos + seg2_len], seg2,
                                      sample_width))
             pos += seg2_len
 
-            if not loop:
-                break
+            # dec times to break our while loop (eventually)
+            times -= 1
 
         output.write(seg1[pos:])
 
         return spawn(data=output)
 
     def append(self, seg, crossfade=100):
-        output = TemporaryFile()
-
         seg1, seg2 = AudioSegment._sync(self, seg)
 
         if not crossfade:
@@ -485,6 +663,8 @@ class AudioSegment(object):
         xf = seg1[-crossfade:].fade(to_gain=-120, start=0, end=float('inf'))
         xf *= seg2[:crossfade].fade(from_gain=-120, start=0, end=float('inf'))
 
+        output = TemporaryFile()
+
         output.write(seg1[:-crossfade]._data)
         output.write(xf._data)
         output.write(seg2[crossfade:]._data)
@@ -520,8 +700,6 @@ class AudioSegment(object):
         if to_gain == 0 and from_gain == 0:
             return self
 
-        frames = self.frame_count()
-
         start = min(len(self), start) if start is not None else None
         end = min(len(self), end) if end is not None else None
 
@@ -548,22 +726,25 @@ class AudioSegment(object):
         # original data - up until the crossfade portion, as is
         before_fade = self[:start]._data
         if from_gain != 0:
-            before_fade = audioop.mul(before_fade, self.sample_width,
+            before_fade = audioop.mul(before_fade,
+                                      self.sample_width,
                                       from_power)
         output.append(before_fade)
 
         gain_delta = db_to_float(to_gain) - from_power
 
         # fades longer than 100ms can use coarse fading (one gain step per ms),
-        # shorter fades will have audible clicks so they use precise fading (one
-        # gain step per sample)
+        # shorter fades will have audible clicks so they use precise fading
+        #(one gain step per sample)
         if duration > 100:
             scale_step = gain_delta / duration
 
             for i in range(duration):
                 volume_change = from_power + (scale_step * i)
                 chunk = self[start + i]
-                chunk = audioop.mul(chunk._data, self.sample_width, volume_change)
+                chunk = audioop.mul(chunk._data,
+                                    self.sample_width,
+                                    volume_change)
 
                 output.append(chunk)
         else:
@@ -582,7 +763,8 @@ class AudioSegment(object):
         # original data after the crossfade portion, at the new volume
         after_fade = self[end:]._data
         if to_gain != 0:
-            after_fade = audioop.mul(after_fade, self.sample_width,
+            after_fade = audioop.mul(after_fade,
+                                     self.sample_width,
                                      db_to_float(to_gain))
         output.append(after_fade)
 
